import mock
import json
import os
import unittest
import shutil
import argparse

from intorods.uxhash.hash import create_hash, hash_dict_to_txt, hash_dict_to_json
from intorods.uxhash.hash import main as hash_main

TEST_OUTPUT_FOLDER = "./test/output_for_tests/"
TEST_INPUT_FOLDER = "./test/input/"


class TestCreateChecksumFile(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        os.makedirs(TEST_OUTPUT_FOLDER)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(TEST_OUTPUT_FOLDER)

    def test_create_hash(self):
        """
        Test the creation of the hashes.
        """
        result = create_hash(TEST_INPUT_FOLDER)
        self.assertEqual(
            result["file1"],
            "9ee1e60c4cf9c254453b240c04a3c563380ff503485a620c55659cdb40aae43c",
        )
        self.assertEqual(
            result["file2"],
            "0f218d4f5147fec04ca763fa4a58e8288b070951e6aa462c691d52bb90671dd9",
        )

    def test_hash_dict_to_txt(self):
        hashdict = {
            "file1": "123123123123",
            "file2": "456456456456",
        }
        result = hash_dict_to_txt(hashdict)
        expected = [
            "123123123123  file1\n456456456456  file2\n",
            "456456456456  file2\n123123123123  file1\n",
        ]
        self.assertIn(result, expected)

    def test_hash_dict_to_json(self):
        hashdict = {
            "file1": "123123123123",
            "file2": "456456456456",
        }
        result = hash_dict_to_json(hashdict)
        result_json = json.loads(result)
        expected = {
            "collection": "",
            "checksum_format": "sha256",
            "version": 1,
            "checksum_encoding": "hex",
            "objects": [
                {"path": "file2", "checksum": "456456456456", "type": "dataobject"},
                {"path": "file1", "checksum": "123123123123", "type": "dataobject"},
            ],
        }
        self.assertCountEqual(result_json["objects"], expected["objects"])

    @mock.patch(
        "argparse.ArgumentParser.parse_args",
        return_value=argparse.Namespace(
            sourcedir=TEST_INPUT_FOLDER,
            output=os.path.join(TEST_OUTPUT_FOLDER, "output_file.json"),
            debuglevel=1,
            procs=1,
            json=True,
            coll="",
        ),
    )
    def test_main(self, mockargs):
        hash_main()
        self.assertTrue(os.path.join(TEST_OUTPUT_FOLDER, "output_file.json"))


if __name__ == "__main__":
    unittest.main()



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))

import filecmp

from intorods.intorods import factory, sync

#
# def sync(sourcefs, sourcepath, destfs, destpath, sfs_name, sfs_opts, dfs_name, dfs_opts,
#         compareChecksums=False, filelist={}, scan=False, worker_processes=1, excludelist=[], 
#         compare=True, minimum_age=0, cs_filters=[], scan_filters=[]):
# 

inputpath = './test/input'


def test_sync_list():
    outputpath = './test/output1'
    fs_source = factory.createfs( 'local' )
    fs_dest = factory.createfs('local')
    if not fs_dest.folderexists(outputpath):
        fs_dest.mkdir(outputpath)
    LIST = { 
        'file1' : '9ee1e60c4cf9c254453b240c04a3c563380ff503485a620c55659cdb40aae43c', 
        'file2': '0f218d4f5147fec04ca763fa4a58e8288b070951e6aa462c691d52bb90671dd9' 
    }
    result = sync(fs_source, inputpath,
            fs_dest, outputpath,
            'local', {}, 
            'local', {},
            compareChecksums=True,
            filelist=LIST, scan=False,
            worker_processes=1, excludelist=[],
            compare=True,
            minimum_age=0, cs_filters=[],
            scan_filters=[]) 
    assert result == True
    assert filecmp.cmp(os.path.join(inputpath, 'file1'), os.path.join(outputpath, 'file1'), shallow=False) == True
    assert filecmp.cmp(os.path.join(inputpath, 'file2'), os.path.join(outputpath, 'file2'), shallow=False) == True

def test_sync_scan():
    outputpath = './test/output2'
    fs_source = factory.createfs( 'local' )
    fs_dest = factory.createfs('local')
    if not fs_dest.folderexists(outputpath):
        fs_dest.mkdir(outputpath)
    result = sync(fs_source, inputpath,
            fs_dest, outputpath,
            'local', {}, 
            'local', {},
            compareChecksums=True,
            scan=True,
            worker_processes=1, excludelist=[],
            compare=True,
            minimum_age=0, cs_filters=[],
            scan_filters=[]) 
    assert result == True
    assert filecmp.cmp(os.path.join(inputpath, 'file1'), os.path.join(outputpath, 'file1'), shallow=False) == True
    assert filecmp.cmp(os.path.join(inputpath, 'file2'), os.path.join(outputpath, 'file2'), shallow=False) == True







import inspect
import json
import os
import sys
import unittest

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from intorods.intorods import (BASECLEAR_SCHEMA_FILE, FILE_FORMAT_BASECLEAR,
                               FILE_FORMAT_TEXT, factory, parse_checksum_file)


class TestParseChecksumFile(unittest.TestCase):
    def test_simple_text_format(self):
        """
        Test the parsing of the simple text-format
        """
        fs_source = factory.createfs( "local" )
        result = parse_checksum_file( fs_source,
                                      fs_source.getfile("./test/data"),
                                      "text_checksumfile1",
                                      FILE_FORMAT_TEXT, 
                                      None )
        self.assertIsNotNone( result )
        self.assertEqual( result["file1"], "123" )
        self.assertEqual( result["file2"], "456" )


    def test_json_format(self):
        """
        Test the parsing of the json format with default path per file
        """
        fs_source = factory.createfs( "local" )
        JSON_SCHEMA=None
        with open( BASECLEAR_SCHEMA_FILE) as f:
            JSON_SCHEMA = json.load(f)
        result = parse_checksum_file( fs_source,
                                      fs_source.getfile("./test/data"),
                                      "json_checksumfile1.json",
                                      FILE_FORMAT_BASECLEAR, 
                                      JSON_SCHEMA )
        self.assertIsNotNone( result )
        self.assertEqual( result["raw_sequences/4711/file1"], "1231231231123123123112312312311231231231123123123112312312311231" )
        self.assertEqual( result["raw_sequences/4711/file2"], "4564564564456456456445645645644564564564456456456445645645644564" )


    def test_json_format_with_path(self):
        """
        Test the parsing of the json format with explicit path per file
        """
        fs_source = factory.createfs( "local" )
        JSON_SCHEMA=None
        with open( BASECLEAR_SCHEMA_FILE) as f:
            JSON_SCHEMA = json.load(f)
        result = parse_checksum_file( fs_source,
                                      fs_source.getfile("./test/data"),
                                      "json_checksumfile2.json",
                                      FILE_FORMAT_BASECLEAR, 
                                      JSON_SCHEMA )
        self.assertIsNotNone( result )
        self.assertEqual( result["any/kind/of/path/file1"], "1231231231123123123112312312311231231231123123123112312312311231" )
        self.assertEqual( result["another/path/file2"], "4564564564456456456445645645644564564564456456456445645645644564" )


    def test_missing_json_format_file(self):
        """
        Test no exception happens when checksum file is not found
        """
        fs_source = factory.createfs( "local" )
        JSON_SCHEMA=None
        with open( BASECLEAR_SCHEMA_FILE) as f:
            JSON_SCHEMA = json.load(f)
        result = parse_checksum_file( fs_source,
                                      fs_source.getfile("./test/data"),
                                      "json_checksumfile3.json",
                                      FILE_FORMAT_BASECLEAR, 
                                      JSON_SCHEMA )
        self.assertIsNone( result )


    def test_globbing_for_checksum_file(self):
        """
        Test file globbing works and will parse file if a SINGLE file matches!
        """
        fs_source = factory.createfs( "local" )
        JSON_SCHEMA=None
        with open( BASECLEAR_SCHEMA_FILE) as f:
            JSON_SCHEMA = json.load(f)
        result = parse_checksum_file( fs_source,
                                      fs_source.getfile("./test/data"),
                                      "json_*file2.json",
                                      FILE_FORMAT_BASECLEAR, 
                                      JSON_SCHEMA )
        self.assertIsNotNone( result )
if __name__ == '__main__':
    unittest.main()



import logging
import math
import multiprocessing as mp
import os
import queue
import re
import time

from intorods.filesys.fs_base import factory
from intorods.filesys.yml_filter import PathFilter

# The sync_worker tries to copy each object max MAX_RETRIES times,
# if comparison fails. After that, the object is added to the error_queue,
# and sync_worker continues with the next object
MAX_RETRIES = 2
# After trying to copy all objects, copying is retried (again) for objects in the
# error queue. This happens at most MAX_ERRQUEUE_RETRIES times
MAX_ERRQUEUE_RETRIES = 2
# MONITOR_INTERVAL determines the frequency at which stats are reported
# and at which the state of worker processes is checked
MONITOR_INTERVAL = 5
# The worker process terminates after MAX_OBJECTS_PER_PROCESS. This
# prevents nemory exhaust on the irods server. The monitor process will
# automatically start a new sync worker if there are still objects to copy
# in the queue
MAX_OBJECTS_PER_PROCESS = 2000

logger = logging.getLogger(__name__)


def sync_worker(process_id, q, sq, error_queue, sfs_name, sfs_opts, dfs_name, dfs_opts, compare, minimum_age):
    """The sync worker syncs items from the queue q
    sfs_name, sfs_opts, dfs_name and dfs_opts contain the source and destination
    filesystem name and connect parameters.
    Items that fail to copy/sync are added to the error queue.
    On succesfull sync, the size of the synced object is added to the 
    statistics queue sq
    """

    def copyobj(entry, dfs, dfile):
        try:
            entry.copyto(dfs, dfile)
            dfs.getfile(dfile).set_mtime(entry.utc_mtime())
        except Exception as ex:
            logger.error('T{}: ERROR COPYING {}. Exception code: {}'.format(
                process_id, entry.path, ex.code if hasattr(ex, "code") else "None"))
            return False
        return True

    source_fs = factory.createfs(sfs_name, **sfs_opts)
    destfs = factory.createfs(dfs_name, **dfs_opts)
    objects_copied = 0
    # Start the copy loop
    while objects_copied < MAX_OBJECTS_PER_PROCESS:
        try:
            # Fetch an item to copy from the queue
            item = q.get(True, 2)
        except queue.Empty:
            logger.debug('T{}: Empty queue: exit'.format(process_id))
            break
        if item is None:
            logger.debug('T{}: Empty queue item: exit'.format(process_id))
            break
        sourcefile, destfile, compareChecksums, checksum = item

        # Get source file object
        try:
            entry = source_fs.getfile(sourcefile)
            if minimum_age:
                now = math.floor(time.time())
                if entry.utc_mtime() >= (now - minimum_age):
                    logger.debug('T{}: File {} is omitted since it was modified shortly.'.format(
                        process_id, sourcefile))
                    break
        except:
            logger.debug('T{}: Error getting source file {}.'.format(
                process_id, sourcefile))
            error_queue.put(item)
            continue

        # Assign checksum from checksum file
        if checksum:
            entry.checksum = checksum

        # Try copying the object MAX_RETRIES times
        retry_counter = 0
        equal = False
        while retry_counter < MAX_RETRIES:
            # Check for existing destination and continue if equal
            try:
                dest_exists = destfs.fileexists(destfile)
            except:
                logger.error('T{}: error querying dest filesystem for {}'.format(
                    process_id, destfile))
                break
            if dest_exists:
                if compare:
                    try:
                        dfile = destfs.getfile(destfile)
                    except:
                        logger.error('T{}: error getting destfile object {}'.format(
                            process_id, destfile))
                        break
                    try:
                        equal = entry.compareto(
                            dfile, compareChecksums=compareChecksums)
                    except Exception as ex:
                        logger.error('T{}: error comparing {} to {}. Exception: {}'.format(
                            process_id, entry.path, destfile, ex))
                        equal = False
                else:
                    # No file comparison required, existence of dest object is enough
                    equal = True

                if equal:
                    sq.put(entry.filesize())
                    break
                else:
                    logger.debug('T{}: objects {} and {} are not equal'.format(
                        process_id, entry.path, destfile))
                    try:
                        destfs.deletefile(destfile)
                    except:
                        logger.error('T{}: error deleting {}'.format(
                            process_id, destfile))

            # Here the actual copy is done.
            # objects_copied is only used to terminate the thread after a certain number of copy actions
            objects_copied += 1
            if not copyobj(entry, destfs, destfile):
                logger.error('T{}: copying {} failed'.format(
                    process_id, entry.path))

            retry_counter += 1

        if not equal:
            logger.error('T{}: copying of {} to {} failed'.format(
                process_id, entry.path, destfile))
            # This is a potential deadlock in case of lots of errors
            # https://stackoverflow.com/questions/59951832/python-multiprocessing-queue-makes-code-hang-with-large-data
            error_queue.put(item)

    # Destroy FS objects
    source_fs.cleanup()
    destfs.cleanup()

    logger.debug('T{}: Finish'.format(process_id))


def sync(sourcefs, sourcepath, destfs, destpath, sfs_name, sfs_opts, dfs_name, dfs_opts,
         compareChecksums=False, filelist={}, scan=False, worker_processes=1, excludelist=[],
         compare=True, minimum_age=0, cs_filters=[], scan_filters=[]):
    # Initialize MP
    job_queue = mp.Queue()
    error_queue = mp.Queue()
    stats_queue = mp.Queue()

    # Create copy processes list
    procs = []

    def new_copy_process(i):
        logger.debug('Creating copy process {}'.format(i))
        p = mp.Process(target=sync_worker, args=(i, job_queue, stats_queue, error_queue,
                                                 sfs_name, sfs_opts, dfs_name, dfs_opts,
                                                 compare, minimum_age))
        p.start()
        return p

    def handle_statistics(ts_start, total_bytes_copied, total_files_copied):
        ts_now = time.time()
        while stats_queue.qsize() > 0:
            bytes_copied = stats_queue.get()
            total_files_copied += 1
            total_bytes_copied += bytes_copied
        byterate = total_bytes_copied/(ts_now-ts_start)
        filerate = total_files_copied/(ts_now-ts_start)
        timeleft = qsize/filerate if filerate > 0 else 0
        avg_size = total_bytes_copied/total_files_copied if total_files_copied > 0 else 0
        logger.info('QUEUE SIZE : {:6d} COPYRATE: {:4.0f}  BYTERATE: {:5.2f} MB/s  AVG SIZE : {:8.2f} kB  TIME: {:.2f} s  ERRORS : {:6d}'.format(
            qsize, filerate, byterate/(1024*1024), avg_size/1024, timeleft, error_queue.qsize()))
        return total_bytes_copied, total_files_copied

    if not destfs.folderexists(destpath):
        destfs.mkdir(destpath)

    ts_start = time.time()

    # Enqueue files to copy
    # copy_jobs is a dict with struct { src_filename : (src_filename, dst_filename, cmp_checksum, checksum_value) }
    # this ensures all source file will be present only once
    copy_jobs = {}

    # Start with scanning the source. since here no cs value will be included
    if scan:
        logger.debug('Queue copy jobs from source fs')
        spf = PathFilter(scan_filters)
        try:
            queue_copyjobs(sourcefs, sourcepath, destfs, destpath, copy_jobs,
                           compareChecksums=compareChecksums, excludelist=excludelist, filter=spf)
        except Exception as ex:
            logger.error(f'ERROR queueing jobs for {sourcepath}: {ex}')
            return False

    if filelist:
        logger.debug('Queue copy jobs from filelist')
        cpf = PathFilter(cs_filters)
        try:
            queue_copyjobs_from_filelist(filelist, sourcepath, destfs, destpath, copy_jobs,
                                         compareChecksums=compareChecksums, excludelist=excludelist, filter=cpf)
        except Exception as ex:
            logger.debug(f'ERROR queueing jobs for {sourcepath}: {ex}')
            return False

    total_files = len(copy_jobs)

    for job in copy_jobs.values():
        job_queue.put(job)

    qsize = job_queue.qsize()
    total_bytes_copied = 0
    total_files_copied = 0

    error_retries = 0
    process_id = 0
    while error_retries < MAX_ERRQUEUE_RETRIES:
        while qsize > 0:
            # Show/handle statistics
            total_bytes_copied, total_files_copied = handle_statistics(
                ts_start, total_bytes_copied, total_files_copied)

            # Handle finished processes
            for p in procs:
                if not p.is_alive():
                    logger.debug('A copy process has ended')
                    p.join()
                    procs.remove(p)
            # Create copy processes if necessary
            desired_processes = min([qsize, worker_processes])
            while len(procs) < desired_processes:
                procs.append(new_copy_process(process_id))
                process_id += 1
            time.sleep(min(MONITOR_INTERVAL, max(qsize//10, 1)))
            qsize = job_queue.qsize()

        while error_queue.qsize() > 0:
            logger.debug('Re-queing {}'.format('x'))
            job_queue.put(error_queue.get())
        qsize = job_queue.qsize()
        error_retries += 1

    logger.debug('Wait for copy jobs to finish')
    errors = 0
    for t in procs:
        # the other side of the deadlock
        # when we reach this point all error_retries have been triggered, anything left in the error_queue is just needed for reporting
        # however the last run is still ongoing, so we can't just join the children,
        # the current implementation polls for a maximum of 100sec per thread, and makes sure the error_queue is regularly read and emptied.
        while True:
            logger.debug(f"waiting to join {t}")
            t.join(100)
            # after timeout check the error queue to prevent deadlock
            while error_queue.qsize() > 0:
                error_queue.get()
                errors += 1
            if not t.is_alive():
                break

    total_bytes_copied, total_files_copied = handle_statistics(
        ts_start, total_bytes_copied, total_files_copied)

    logger.info('Had {} files to sync, handled {}'.format(
        total_files, total_files_copied))
    logger.info('Files with sync error: {}'.format(errors))
    success = (errors == 0) and (total_files_copied == total_files)

    return success


def queue_copyjobs_from_filelist(filelist, sourcepath, destfs, destpath, copy_jobs,
                                 compareChecksums=True, excludelist=[], filter=None):
    dircache = []
    copy_counter = 0

    exclude_relist = [re.compile(exclude, re.IGNORECASE)
                      for exclude in excludelist]

    for filename in filelist:
        if filter and (not filter.isFileIncluded(filename)):
            continue

        # Skip files that match one of the exclude patterns
        match = False
        for exclude_re in exclude_relist:
            if exclude_re.match(filename):
                match = True
        if match:
            continue
        #
        destfile = os.path.join(destpath, filename)
        destdir = os.path.dirname(destfile)
        # Create the destination directory, if non-existing
        if not destdir in dircache:
            if not destfs.folderexists(destdir):
                destfs.mkdir(destdir)
            dircache.append(destdir)
        # Add files to the copy queue
        src_filename = os.path.join(sourcepath, filename)
        copy_jobs[src_filename] = (src_filename, os.path.join(destpath, filename),
                                   compareChecksums, filelist[filename])
        copy_counter += 1
    return copy_counter


def queue_copyjobs(sourcefs, sourcebase, destfs, destbase, copy_jobs,
                   compareChecksums=False, excludelist=[], relpath='', filter=None):

    #
    # PATH variable example
    #
    #  /data/vir-ont/20201204_N2-MIN83-oke/20201204_1704_X1_FAO64587_bf55c64/fast5_pass/FAO64587_pass_0ad68019_3146.fast5
    # |                                sourcebase                           | relpath  |
    # |                                    sourcepath                                  |
    # |                                                                     |           relfilepath                      |
    #                                      |   parentDir                    |
    #
    #  /demoZone/projects/nsglab/minion/20201204_1704_X1_FAO64587_bf55c64/fast5_pass/FAO64587_pass_0ad68019_3146.fast5
    # |                                 destbase                         | relpath  |
    # |                                                                  |           relfilepath                      |
    # |                                        destdir                              |

    dircache = []
    copy_counter = 0

    exclude_relist = [re.compile(exclude, re.IGNORECASE)
                      for exclude in excludelist]

    sourcepath = os.path.join(sourcebase, relpath)

    # I guess only intorods knows what the base directory was sync is called for.
    #parentdir = sourcebase.split(os.sep)[-1]

    for entry in sourcefs.lsdirnames(sourcepath):
        relsubdir = os.path.join(relpath, entry)
        copy_counter += queue_copyjobs(sourcefs, sourcebase, destfs, destbase, copy_jobs,
                                       compareChecksums=compareChecksums, excludelist=excludelist, relpath=relsubdir, filter=filter)

    for entry in sourcefs.lsfilenames(sourcepath):

        relfilepath = os.path.join(relpath, entry)
        # Skip files that match one of the exclude patterns
        if filter and (not filter.isFileIncluded(relfilepath)):
            continue
        match = False
        for exclude_re in exclude_relist:
            if exclude_re.match(relfilepath):
                match = True
        if match:
            continue

        destdir = os.path.join(destbase, relpath).rstrip('/')
        if not destdir in dircache:
            if not destfs.folderexists(destdir):
                destfs.mkdir(destdir, parents=True)
            dircache.append(destdir)

        src_filename = os.path.join(sourcebase, relfilepath)
        copy_jobs[src_filename] = (src_filename, os.path.join(
            destbase, relfilepath), compareChecksums, None)
        copy_counter += 1

    return copy_counter



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 25 11:00:00 2023

@author: jansser
"""

import os
import stat
from datetime import datetime

import ftputil

from intorods.filesys.fs_base import factory, fs_base, fsobject_base


class file_ftp(fsobject_base):
    def __init__(self, fso, path):
        super().__init__(fso, path)
        self._lstat = self.fso.ftp.lstat(self.path)

    def filesize(self):
        return self.fso.ftp.path.getsize(self.path)

    def isfile(self):
        return not stat.S_ISDIR(self._lstat.st_mode)


    def accessible(self):
        if self.isdir():
            try:
                self.fso.ftp.listdir(self.path)
            except:
                return False
            return True
        else:
            try:
                fo = self.fso.ftp.file(self.path, "r")
                if fo:
                    return fo.readable()
            except IOError:
                return False
        return False

    def open(self, mode):
        return self.fso.ftp.open(self.path, mode)

    def set_mtime(self, mtime):
        # not implemented in ftputil, using raw FTP via underlying ftplib 
        # this is not implemented on all FTP servers, or you may lack rights!
        formatted_time = datetime.utcfromtimestamp(mtime).strftime('%Y%m%d%H%M%S')
        self.ftp._session.voidcmd(f"MFMT {self.path} {formatted_time}")

    def utc_mtime(self):
        return self._lstat.st_mtime

    def isdir(self):
        return stat.S_ISDIR(self._lstat.st_mode)

class fs_ftp(fs_base):
    def __init__(self, hostname='', user='anonymous', password='me@domain.com'):
        super().__init__(supportsopen=True)
        self.hostname = hostname
        self.user = user
        self.ftp = ftputil.FTPHost(hostname, user, password)
        # try to correct the difference in timezone. Due to the implementation,
        # this is only possible if you have write access to the ftp server!
        try:
            self.ftp.synchronize_times()
        except:
            pass

    @staticmethod
    def factory(**kwargs):
        return fs_ftp(**kwargs)

    def fileexists(self, path):
        try:
            fileobject = self.getfile(path)
        except:
            return False
        return fileobject.isfile()

    def folderexists(self, path):
        try:
            self.ftp.exists(path)
        except:
            return False
        return True

    def _getfile(self, path):
        return file_ftp(self, path)

    def ls(self, path):
        result = []
        try:
            for file in self.ftp.listdir(path):
                fullpath = os.path.join(path, file)
                fileobject = self.getfile(fullpath)
                result.append(fileobject)
        except:
            pass
        return result

    def mkdir(self, path, parents=False):
        exit(1)

    def open(self, path, mode):
        return self.ftp.open(path, mode)


factory.register('ftp', fs_ftp.factory,
                 'hostname=<hostname>,user=<username>,password=<password>')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 19 16:08:06 2019

@author: Erwin van Wieringen
"""

import os
import re
import stat
import warnings

from cryptography.utils import CryptographyDeprecationWarning

with warnings.catch_warnings():
    warnings.filterwarnings('ignore', category=CryptographyDeprecationWarning)
    import paramiko

from intorods.filesys.fs_base import factory, fs_base, fsobject_base


def esc(mystr):
    return re.sub('([$])', r'\\\1', mystr).strip()


class file_scp(fsobject_base):
    def __init__(self, fso, path):
        super().__init__(fso, path)
        self._lstat = self.fso.sftp.lstat(self.path)

    def calculate_checksum(self):
        myhash = ""
        try:
            I = self.open("r")
            myhash = I.check("sha256")
        except IOError:
            pass
        if myhash == "":
            command = 'sha256sum "' + esc(self.path) + '"  | cut -f 1 -d" " '
            stdin, stdout, stderr = self.fso.ssh.exec_command(command)
            out0 = stdout.read().splitlines()[0]
            myhash = out0.decode().split("'")[0]
        return myhash

    def filesize(self):
        # return self.fso.sftp.lstat(self.path).st_size
        return self._lstat.st_size

    def isfile(self):
        return not stat.S_ISDIR(self._lstat.st_mode)

    def open(self, mode):
        return self.fso.sftp.open(self.path, mode)

    def set_mtime(self, mtime):
        self.fso.sftp.utime(self.path, (mtime, mtime))

    def utc_mtime(self):
        return self._lstat.st_mtime

    def isdir(self):
        return stat.S_ISDIR(self._lstat.st_mode)


class fs_scp(fs_base):
    def __init__(self, hostname='', user='', private_key=None, public_key='', port=22):
        super().__init__(supportsopen=True)
        self.hostname = hostname
        self.user = user if user else os.environ('USER')
        self.port = port

        self.ssh = paramiko.SSHClient()
        self.ssh.load_system_host_keys()
        self.ssh.set_missing_host_key_policy(paramiko.WarningPolicy)
        self.ssh.connect(self.hostname, username=self.user, port=self.port, key_filename=private_key)
        self.sftp = paramiko.SFTPClient.from_transport(self.ssh.get_transport())


    @staticmethod
    def factory(**kwargs):
        return fs_scp(**kwargs)

    def fileexists(self, path):
        try:
            fileobject = self.getfile(path)
        except:
            return False
        return fileobject.isfile()

    def folderexists(self, path):
        try:
            self.sftp.listdir(path)
        except:
            return False
        return True

    def _getfile(self, path):
        return file_scp(self, path)

    def ls(self, path):
        result = []
        for file in self.sftp.listdir(path):
            fullpath = path + '/' + file
            fileobject = self.getfile(fullpath)
            result.append(fileobject)
        return result

    def mkdir(self, path, parents=False):
        exit(1)

    def open(self, path, mode):
        return self.sftp.open(path, mode)


factory.register('scp', fs_scp.factory, 'hostname=<hostname>,user=<username>,private_key=<private key file>')



import unittest

from intorods.filesys.yml_filter import PathFilter


class TestYmlFilter(unittest.TestCase):

    def test_no_filter(self):
        filter = []
        pf = PathFilter( filter )
        f = pf.isFileIncluded( "path/to/file.txt")
        self.assertEqual( f, True, "")
        d = pf.isDirIncluded( "path/to/somewhere" )
        self.assertEqual( d, True, "")

    def test_all_excluded(self):
        filter = [ "- /"]
        pf = PathFilter( filter )
        pf.show()
        f = pf.isFileIncluded( "path/to/file.txt")
        self.assertEqual( f, False, "")
        d = pf.isDirIncluded( "path/to/somewhere" )
        self.assertEqual( d, False, "")

    def test_all_included(self):
        filter = [ "+ /"]
        pf = PathFilter( filter )
        pf.show()
        f = pf.isFileIncluded( "path/to/file.txt")
        self.assertEqual( f, True, "")
        f = pf.isFileIncluded( "file.txt")
        self.assertEqual( f, True, "")
        d = pf.isDirIncluded( "path/to/somewhere" )
        self.assertEqual( d, True, "")

    def test_exclude_dirs(self):
        filter = [ "+ /", "- path/to/"]
        pf = PathFilter( filter )
        pf.show()
        f = pf.isFileIncluded( "path/to/file.txt")
        self.assertEqual( f, False, "")
        d = pf.isDirIncluded( "path/to/somewhere" )
        self.assertEqual( d, False, "")

    def test_include_dirs(self):
        filter = [ "- /", "+ A/B/" ]
        pf = PathFilter( filter )
        pf.show()
        d = pf.isDirIncluded("A/B")
        self.assertEqual(d, True, "")
        d = pf.isDirIncluded("A/B/C")
        self.assertEqual(d, True, "")


    def test_exclude_dirs_include_files(self):
            filter = [ "- /", 
                    "+ A/A2/A3/ .*\.in",
                    "- A/A2/A3/ .*\.ex",
                    "+ B/", 
                    "- B/B2/"]
            pf = PathFilter( filter )
            pf.show()
            f = pf.isFileIncluded( "file.txt")
            self.assertEqual( f, False, "")
            f = pf.isFileIncluded( "A/file.txt")
            self.assertEqual( f, False, "")
            f = pf.isFileIncluded( "A/A2/file.txt")
            self.assertEqual( f, False, "")
            f = pf.isFileIncluded( "A/A2/A3/file.txt")
            self.assertEqual( f, False, "")
            f = pf.isFileIncluded( "A/A2/A3/file.in")
            self.assertEqual( f, True, "")
            f = pf.isFileIncluded( "A/A2/A3/file.ex")
            self.assertEqual( f, False, "")
            f = pf.isFileIncluded( "B/file.txt")
            self.assertEqual( f, True, "")
            f = pf.isFileIncluded( "B/X/file.txt")
            self.assertEqual( f, True, "")
            f = pf.isFileIncluded( "BV/A/file.txt")
            self.assertEqual( f, False, "")


    def test_exclude_dirs_include_files_opposite(self):
            filter = [ "+ /", 
                    "- A/A2/A3/ .*\.in",
                    "+ A/A2/A3/ .*\.ex",
                    "- B/", 
                    "+ B/B2/"]
            pf = PathFilter( filter )
            pf.show()
            f = pf.isFileIncluded( "file.txt")
            self.assertEqual( f, True, "")
            f = pf.isFileIncluded( "A/file.txt")
            self.assertEqual( f, True, "")
            f = pf.isFileIncluded( "A/A2/file.txt")
            self.assertEqual( f, True, "")
            f = pf.isFileIncluded( "A/A2/A3/file.txt")
            self.assertEqual( f, True, "")
            f = pf.isFileIncluded( "A/A2/A3/file.in")
            self.assertEqual( f, False, "")
            f = pf.isFileIncluded( "A/A2/A3/file.ex")
            self.assertEqual( f, True, "")
            f = pf.isFileIncluded( "B/file.txt")
            self.assertEqual( f, False, "")
            f = pf.isFileIncluded( "B/X/file.txt")
            self.assertEqual( f, False, "")
            f = pf.isFileIncluded( "B/B2A/file.txt")
            self.assertEqual( f, False, "")

    def test_excl_dir_incl_file(self):
        filter = [ "- /", 
                   "- A/A2/A3/",
                    "+ A/A2/A3/ .*\.in"
                    ]
        pf = PathFilter( filter )
        pf.show()
        f = pf.isFileIncluded( "A/A2/A3/file.txt")
        self.assertEqual( f, False, "")
        f = pf.isFileIncluded( "A/A2/A3/file.in")
        self.assertEqual( f, True, "")
        
    def test_root_files(self):
        filter = [] # TODO fill in the correct filter for this test
        pf = PathFilter( filter )
        pf.show()
        # f = pf.isFileIncluded( "file.txt")
        # self.assertEqual( f, False, "")
        # f = pf.isFileIncluded( "file.in")
        # self.assertEqual( f, True, "")

    def test_absolute_paths(self):
        filter = [ "- /", 
                   "+ /A/"
                 ]
        pf = PathFilter( filter )
        pf.show()
        f = pf.isFileIncluded( "A/file.txt")
        self.assertEqual( f, False, "")
        f = pf.isFileIncluded( "/A/file.in")
        self.assertEqual( f, True, "")

    def test_base_calls(self):
        filter = [ "+ /",
            "- Data/Intensities/BaseCalls/",
            "+ Data/Intensities/BaseCalls/ .*" ]
        pf = PathFilter( filter )
        pf.show()
        d = pf.isDirIncluded( "Data/Intensities/BaseCalls/SubDir")
        self.assertEqual( d, False, "")
        d = pf.isDirIncluded( "Data/Intensities/BaseCalls/Phasing")
        self.assertEqual( d, False, "")
        f = pf.isFileIncluded( "Data/Intensities/BaseCalls/Phasing/test.txt")
        self.assertEqual( f, False, "")
        d = pf.isDirIncluded( "Data/Intensities/BaseCalls/Alignment")
        self.assertEqual( d, False, "")
        f = pf.isFileIncluded( "Data/Intensities/BaseCalls/Alignment/test.txt")
        self.assertEqual( f, False, "")
        f = pf.isFileIncluded( "Data/Intensities/BaseCalls/11_S11_L001_R1_001.fastq.gz")
        self.assertEqual( f, True, "")
        f = pf.isFileIncluded( "Data/Intensities/BaseCalls/12_S12_L001_R1_001.fastq.gz")
        self.assertEqual( f, True, "")

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 19 16:08:06 2019

@author: Erwin van Wieringen
"""

import glob
import hashlib
import os
import re
import stat
import warnings

from cryptography.utils import CryptographyDeprecationWarning

with warnings.catch_warnings():
    warnings.filterwarnings('ignore', category=CryptographyDeprecationWarning)
    import paramiko

from intorods.filesys.fs_base import factory, fopen, fs_base, fsobject_base


def esc(mystr):
    return re.sub('([$])', r'\\\1', mystr).strip()


class file_sftp(fsobject_base):
    def __init__(self, fso, path):
        super().__init__(fso, path)
        self._lstat = self.fso.sftp.lstat(self.path)

    def filesize(self):
        # return self.fso.sftp.lstat(self.path).st_size
        return self._lstat.st_size

    def isfile(self):
        return not stat.S_ISDIR(self._lstat.st_mode)

    def accessible(self):
        if self.isdir():
            try:
                self.fso.sftp.listdir(self.path)
            except:
                return False
            return True
        else:
            try:
                fo = self.fso.sftp.file(self.path, "r")
                if fo:
                    return fo.readable()
            except IOError:
                return False
        return False

    def calculate_checksum(self):
        BUF_SIZE = 1024 * 1024
        sha256 = hashlib.sha256()
        with fopen(self, 'r') as f:
            while True:
                data = f.read(BUF_SIZE)
                if not data:
                    break
                sha256.update(data)
        return sha256.hexdigest()

    def open(self, mode):
        return self.fso.sftp.open(self.path, mode)

    def set_mtime(self, mtime):
        self.fso.sftp.utime(self.path, (mtime, mtime))

    def utc_mtime(self):
        return self._lstat.st_mtime

    def isdir(self):
        return stat.S_ISDIR(self._lstat.st_mode)


class fs_sftp(fs_base):
    def __init__(self, hostname='', user='', password=''):
        super().__init__(supportsopen=True)
        self.hostname = hostname
        self.user = user
        port = 22
        self.t = paramiko.Transport((self.hostname, port))
        self.t.connect(username=user, password=password)

#        transport.start_client(event=None, timeout=15)
#        transport.get_remote_server_key()
#        transport.auth_publickey(username, sftp_key, event=None)
#        transport.auth_password(username, password, event=None)
#        sftp = paramiko.SFTPClient.from_transport(transport)
        self.sftp = paramiko.SFTPClient.from_transport(self.t)

    @staticmethod
    def factory(**kwargs):
        return fs_sftp(**kwargs)

    def fileexists(self, path):
        try:
            fileobject = self.getfile(path)
        except:
            return False
        return fileobject.isfile()

    def folderexists(self, path):
        try:
            self.sftp.listdir(path)
        except:
            return False
        return True

    def _getfile(self, path):
        return file_sftp(self, path)

    def ls(self, path):
        result = []
        try:
            for file in self.sftp.listdir(path):
                fullpath = path + '/' + file
                fileobject = self.getfile(fullpath)
                result.append(fileobject)
        except:
            pass
        return result

    def lsfilenames(self, path):
        return self.sftp.listdir(path)

    def mkdir(self, path, parents=False):
        exit(1)

    def open(self, path, mode):
        return self.sftp.open(path, mode)


factory.register('sftp', fs_sftp.factory,
                 'hostname=<hostname>,user=<username>,password=<password>')



"""
Created on Mon May  6 14:28:12 2019
@author: Erwin van Wieringen
Purpose: abstract filesystem base class
"""

import fnmatch
import hashlib
import logging
import os
import shutil
from abc import ABC, abstractmethod
from datetime import datetime

BUF_SIZE = 1024 * 1024 * 4

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class fsFactory():
    _fscreators = {}

    def __init__(self):
        self._fscreators = {}
        self._optiontext = {}

    def register(self, name, creator, optionText):
        self._fscreators[name] = creator
        self._optiontext[name] = optionText

    def createfs(self, name, **params):
        creator = self._fscreators[name]
        return creator(**params)

    def listfs(self):
        print('Filesystem   Options\n===============================================')
        for i in self._fscreators:
            print("%-10s %-40s" % (i, self._optiontext[i]))


factory = fsFactory()


# class fsbuilder():
#    def __init__(self):
#        self._instance = None
#
#    def options(self):
#        return '-'


class fsobject_base(ABC):
    def __init__(self, fso, path):
        self.fso = fso
        self.path = path
        self._checksum = None

    @abstractmethod
    def isdir(self):
        pass

    @abstractmethod
    def isfile(self):
        pass

    def accessible(self):
        # Test if object is accessible for reading
        # can be reimplemented in derived classes
        try:
            _ = self.open('r')
        except:
            return False
        return True

    def calculate_checksum(self):
        #        logger.debug('Expensive checksum calc for %s' % self.path)
        sha256 = hashlib.sha256()
        with fopen(self, 'rb') as f:
            while True:
                data = f.read(BUF_SIZE)
                if not data:
                    break
                sha256.update(data)
        return sha256.hexdigest()

    @property
    def checksum(self):
        if self._checksum is None:
            self._checksum = self.calculate_checksum()
        return self._checksum

    @checksum.setter
    def checksum(self, cs):
        self._checksum = cs

    def compareto(self, other, compareChecksums=True):
        if compareChecksums:
            if self.checksum == other.checksum:
                return True
                # fixes an issue where files are registered with correct hash in ICAT, but not copied correctly
                if self.filesize() == other.filesize():
                    if self.utc_mtime() == other.utc_mtime():
                        return True
        else:
            if self.filesize() == other.filesize():
                if self.utc_mtime() == other.utc_mtime():
                    return True
        return False

    def copyto(self, dest_fs, dest_path):
        source_fp = self.open('rb')
        dest_fp = dest_fs.createfile(dest_path)
        shutil.copyfileobj(source_fp, dest_fp, BUF_SIZE)
        source_fp.close()
        dest_fp.close()
        dest_fs.invalidate_cache_entry(dest_path)

    @abstractmethod
    def filesize(self):
        pass

    def open(self, mode):
        raise 'fsobject_base.open not implemented'

    def shortname(self):
        return self.path.split('/')[-1]

    @abstractmethod
    def utc_mtime(self):
        pass


class fopen():
    def __init__(self, ff, mode):
        self.objf = ff
        self.mode = mode
        self.fp = None

    def __enter__(self):
        self.fp = self.objf.open(self.mode)
        return self.fp

    def __exit__(self, ftype, value, traceback):
        self.fp.close()


class fs_base(ABC):
    @abstractmethod
    def __init__(self, supportsopen=False):
        self.supportsopen = supportsopen
        self.files = {}

    def refresh(self):
        # Can be used to reconnect in the case of filesystems that are
        # connection-based (smb, irods, ftp)
        pass

    def cleanup(self):
        pass

    def invalidate_cache(self):
        self.files = {}

    def invalidate_cache_entry(self, path):
        if path in self.files:
            del self.files[path]

    @abstractmethod
    def ls(self, path):
        pass

    def lsdirnames(self, path):
        result = [a.shortname() for a in self.ls(path) if a.isdir()]
        return result

    def lsfilenames(self, path):
        result = [a.shortname() for a in self.ls(path) if not a.isdir()]
        return result

    def lsdirs(self, path, skip_inaccessible=False):
        if skip_inaccessible:
            result = [a for a in self.ls(
                path, skip_inaccessible=True) if a.isdir() and a.accessible()]
        else:
            result = [a for a in self.ls(path) if a.isdir()]
        return result

    def _pathsplit(self, path):
        spath = path.split('/')
        base = '/' + '/'.join(spath[1:-1])
        return base, spath[-1]

    @abstractmethod
    def fileexists(self, path):
        pass

    def glob(self, path):
        result = []
        head, tail = os.path.split(path)
        for filename in self.lsfilenames(head):
            if fnmatch.fnmatch(filename, tail):
                result.append(os.path.join(head, filename))
        return result

    def foldermtime(self, path):
        mtime = 0
        pathListing = self.ls(path)
        for entry in pathListing:
            if entry.isdir() and entry.shortname() != '.' and entry.shortname() != '..':
                mtime = max(mtime, self.foldermtime(entry.path))
            else:
                mtime = max(mtime, entry.utc_mtime())
        return mtime

    @abstractmethod
    def folderexists(self, path):
        pass

    @abstractmethod
    def _getfile(self, path):
        pass
        # return fsobject_base(self, path)

    def getfile(self, path):
        if path not in self.files:
            self.files[path] = self._getfile(path)
        return self.files[path]

    def getfolder(self, path):
        return self.getfile(path)

    @abstractmethod
    def mkdir(self, path, parents=False):
        pass

    def createfile(self, path):
        logger.error('%s :createfile not implemented in ' % self.__class__)
        exit(2)

    def deletefile(self, path):
        logger.error('%s :deletefile not implemented in ' % self.__class__)
        exit(2)

    def syncfolderto(self, sourcepath, destfs, destpath):
        logger.debug("sync_folder %s to %s" % (sourcepath, destpath))

        copyCounter = 0

        if not destfs.folderexists(destpath):
            destfs.mkdir(destpath)

        for entry in self.ls(sourcepath):
            destfile = os.path.join(destpath, entry.shortname())
            if entry.isdir():
                if not entry.shortname() in ['.', '..']:
                    self.syncfolderto(entry.path, destfs, destfile)
            else:
                bcopy = False
                if not destfs.fileexists(destfile):
                    logger.debug('Destination not found')
                    bcopy = True
                elif not entry.compareto(destfs.getfile(destfile)):
                    logger.debug('Comparison failed')
                    bcopy = True
                if bcopy:
                    logger.debug('Copy %s to %s' % (entry.path, destfile))
                    copyCounter += 1
                    if destfs.supportsopen:
                        entry.copyto(destfs, destfile)
                    elif self.supportsopen():
                        with entry.open("rb") as fp_source:
                            destfs.copyfrom(fp_source, destfile)
                    else:
                        logger.error('Error: cannot copy %s to %s' %
                                     (entry.shortname(), destfile))
                        exit(2)

                    destfs.getfile(destfile).set_mtime(entry.utc_mtime())

        return copyCounter



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May  8 10:25:59 2019

@author: Erwin van Wieringen
"""

import glob
import math
import os

from intorods.filesys.fs_base import factory, fs_base, fsobject_base


class file_local(fsobject_base):
    def __init__(self, fso, path):
        super().__init__(fso, path)
        self.fp = None
        self.__update()

    def __update(self):
        self.stat = os.stat(self.path)

    def filesize(self):
        return self.stat.st_size

    def isdir(self):
        return os.path.isdir(self.path)

    def isfile(self):
        return os.path.isfile(self.path)

    def accessible(self):
        if self.isdir():
            return os.access(self.path, os.X_OK)
        else:
            return os.access(self.path, os.R_OK)

    def close(self):
        self.fp.close()

    def open(self, mode):
        self.fp = open(self.path, mode)
        return self.fp

    def utc_mtime(self):
        return math.floor(self.stat.st_mtime)

    def set_mtime(self, mtime):
        os.utime(self.path, (mtime, mtime))
        self.__update()


class fs_local(fs_base):
    def __init__(self):
        super().__init__(supportsopen=True)

    def createfile(self, path):
        return open(path, 'wb')

    @staticmethod
    def factory(**kwargs):
        return fs_local(**kwargs)

    def fileexists(self, path):
        # Check existence of path. Path may contain glob wildcard characters.
        return len(glob.glob(str(path))) > 0

    def folderexists(self, path):
        return os.path.isdir(path)

    def _getfile(self, path):
        return file_local(self, path)

    def getfolder(self, path):
        return file_local(self, path)

    def ls(self, path, skip_inaccessible=False):
        result = []
        for entry in os.listdir(path):
            fullpath = path + '/' + entry
            try:
                ff = self.getfile(fullpath)
                result.append(ff)
            except Exception as e:
                if not skip_inaccessible:
                    raise e
        return result

    def glob(self, path):
        return glob.glob(str(path))
        pass

    def lsdirnames(self, path):
        result = [f.name for f in os.scandir(path) if f.is_dir()]
        return result

    def lsfilenames(self, path):
        result = [f.name for f in os.scandir(path) if not f.is_dir()]
        return result

    def mkdir(self, path, parents=False):
        if parents:
            os.makedirs(path)
        else:
            os.mkdir(path)

    def open(self, path, mode):
        return open(path, mode)

    def deletefile(self, path):
        os.remove(path)


factory.register('local', fs_local.factory, '')



#!/usr/bin/env python3

"""
Created on Mon May  6 14:28:12 2019
@author: Erwin van Wieringen

"""

import argparse
import hashlib
import json
import logging
import logging.config
import os
import re
import sys
import time

import irods.exception as ex
import yaml
from jsonschema import ValidationError, validate

import intorods.filesys.fs_ftp
import intorods.filesys.fs_irods
import intorods.filesys.fs_local
import intorods.filesys.fs_scp
import intorods.filesys.fs_sftp

from .filesys.fs_base import factory
from .filesys.fs_smb import wintoux
from .filesys.loghandlers import log_to_stdout, log_to_syslog
from .filesys.sync import sync

logger = logging.getLogger(None)
for _ in ("irods.connection", "irods.manager.metadata_manager", "irods.message", "irods.pool", 
          "irods.session", "paramiko.transport.sftp", "SMB.SMBConnection"):
    logging.getLogger(_).disabled = True
logger.setLevel(logging.WARNING)

FILE_FORMAT_TEXT = 'text'

#the baseclear format
FILE_FORMAT_BASECLEAR = 'baseclear'
BASECLEAR_SCHEMA_FILE = 'schemas/baseclear_checksum_file_schema.json'

FILE_FORMAT_GENERIC_JSON = 'generic'
GENERIC_JSON_SCHEMA_FILE = 'schemas/pipeline_output_schema.json'

METADATA_TEMPLATE = {
    '%sf': lambda s, d: os.path.basename(s),
    '%sp': lambda s, d: s,
    '%df': lambda s, d: os.path.basename(d),
    '%dp': lambda s, d: d,
    '%t': lambda s, d: str(time.time())
}


def formattime(seconds):
    """
    Creates a readable datetime from a duration in seconds
    """
    days = (seconds // 86400)
    hours = (seconds // 3600) % 24
    minutes = (seconds // 60) % 60
    seconds = seconds % 60
    out = "%2i days %2i:%2i:%2i " % (days, hours, minutes, seconds)
    return out


def parse_extra_options(options):
    """
    Creates a dict from a list of options in the form
    name1=value1,name2=value2
    """
    outp = {}
    if options:
        field = options.split(',')
        for option in field:
            kwval = option.split('=')
            outp[kwval[0]] = kwval[1]
    return outp


def parse_checksum_file(fs_source, folder, checksumfile, file_format, json_schema):
    """
        returns a dictionary, containing complete/path/to/file --> hash-value
        returns None to indicate to caller the checksumfile is missing / invalid, either abort or continue with next folder...
    """
    cslist = {}
    logger.debug('Searching checksums-file %s' % checksumfile)
    checksumfilepattern = os.path.join(folder.path, checksumfile)
    csfile_list = fs_source.glob(checksumfilepattern)
    if not 1 == len(csfile_list):
        logger.debug('No definitive match of checksumfile found. Pattern: {} in folder: {}, matches: {}'.format(
            checksumfile, folder.path, len(csfile_list)))
        return None
    checksumfilepath = csfile_list[0]
    if not fs_source.fileexists(checksumfilepath):
        logger.error('Required checksumfile %s not found' % checksumfile)
        return None
    logger.debug('Adding checksums from file %s' % checksumfilepath)
    csfile = fs_source.getfile(checksumfilepath)
    if not csfile.accessible():
        logger.error('Checksumfile %s is not accessible' % checksumfilepath)
        return None
    #put also the samplesheet itself into the cslist for synchronization
    csf_name = os.path.basename(checksumfilepath)
    cslist[wintoux(csf_name)] = None
    #depending on file format, change behavior....
    if file_format == FILE_FORMAT_TEXT:
        with csfile.open('r') as fh:
            lines = fh.readlines()
            for line in lines:
                if not type(line) is str:
                    line = line.decode('utf-8')
                hashvalue, filename = line.rstrip().split('  ')
                cslist[wintoux(filename)] = hashvalue
        return cslist

    elif file_format == FILE_FORMAT_GENERIC_JSON:
        with csfile.open('r') as fh:
            csf_content = json.load(fh)
            try:
                validate(instance=csf_content, schema=json_schema)
            except ValidationError as ex:
                logger.error("ERROR validating checksumfile '{}' threw: '{}'".format(
                    checksumfilepath, ex.message))
                return None
            #parse content
            for data_object in csf_content['objects']:
                hashvalue = data_object['checksum']
                filepath = wintoux(data_object['path'])
                objtype = data_object['type']
                if objtype != 'dataobject':
                    continue
                if filepath in cslist:
                    logger.error(
                        "ERROR parsing checksumfile '{}' is referenced multiple times".format(filepath))
                    return None
                #further sanity checks?
                cslist[filepath] = hashvalue
        return cslist

    elif file_format == FILE_FORMAT_BASECLEAR:
        #magic string, sub directory the actual seqencer files are contained (within further directory structures)
        #Better would be to define this folder in a field of the json-file
        #TODO: once BaseClear abides by the new format, remove magic strings and use always path-property
        SEQUENCE_FILE_DIR = 'raw_sequences'
        #validate against schema definition
        with csfile.open('r') as fh:
            csf_content = json.load(fh)
            try:
                validate(instance=csf_content, schema=json_schema)
            except ValidationError as ex:
                logger.error("ERROR validating checksumfile '{}' threw: '{}'".format(
                    checksumfilepath, ex.message))
                return None
            #parse content
            for sample in csf_content['samples']:
                flowcellid = sample['flowcellid']
                for pair in sample['pairs']:
                    if 'path' in pair.keys():
                        path = pair['path']
                    else:
                        path = os.path.join(SEQUENCE_FILE_DIR, flowcellid)
                    hashvalue = pair['checksum']
                    filename = pair['filename']
                    filepath = os.path.join(path, filename)
                    ux_filepath = wintoux(filepath)
                    if ux_filepath in cslist:
                        logger.error("ERROR parsing checksumfile '{}', file: '{}' is referenced multiple times".format(
                            checksumfilepath, ux_filepath))
                        return None
                    #further sanity checks?
                    cslist[ux_filepath] = hashvalue
        return cslist
    else:
        logger.error('Unknown file format "%s" for checksum file!' %
                     file_format)
        return None


def resolve_meta_template(value, sourcefolder, destfolder):
    """Replace template variables in metadata:
        %sf  basename of source path
        %sp  full source path
        %df  basename of dest path
        %dp  full dest path
    """

    for template, func in METADATA_TEMPLATE.items():
        value = value.replace(template, func(sourcefolder, destfolder))
    return value


def test_completion_avu(dfolder, completion_avu):
    """Will return true if completion avu is present on dfolder
    """
    completion_attr, completion_value = completion_avu.split('=')
    try:
        value = dfolder.getmeta(completion_attr)
    except KeyError:
        return False

    return completion_value == value


def sync_folder(fs_source, folder, fs_dest, destfolder, objdfolder,
                sfs_name, sfs_opts, dfs_name, dfs_opts,
                compareChecksums,
                cslist,
                metadata,
                copy_procs,
                excludelist,
                compare,
                minimum_age,
                cs_filters,
                scan_filters,
                scan):
    logger.debug('Replicating to irods folder {}'.format(objdfolder.path))
    syncresult = sync(fs_source, folder.path, fs_dest, destfolder,
                      sfs_name, sfs_opts, dfs_name, dfs_opts,
                      compareChecksums=compareChecksums,
                      filelist=cslist, worker_processes=copy_procs,
                      excludelist=excludelist,
                      compare=compare,
                      minimum_age=minimum_age,
                      scan_filters=scan_filters,
                      cs_filters=cs_filters,
                      scan=scan)
    if syncresult:
        logger.info('Folders are EQUAL')
        # Add metadata from metadata list
        for attr, v in metadata.items():
            value = resolve_meta_template(v, folder.path, destfolder)
            objdfolder.replaceorsetmeta(attr, value)
    else:
        logger.error('SYNC FAILED for folder %s' % folder.path)
    return syncresult


def replicate_data_folder(sfs_name, sfs_opts, dfs_name, dfs_opts,
                          sourcepath, destpath,
                          checksumfile='', checksumfileformat='', checksumfileschema='', compareChecksums=True, flagfile='', flag_age=0,
                          minage=0, maxage=999999999, pattern='',
                          metadata={}, skip_subdirs=0,
                          copy_procs=1, excludelist=[],
                          completion_avu=None,
                          mark_complete=False,
                          compare=True,
                          timestamp_list=[],
                          minimum_age=None,
                          cs_filter_file='',
                          scan=False,
                          scan_filter_file=''):
    """Replicate a data folder from source filesystem to destination 
    file system by iterating over the direct subfolders of the given
    sourcepath and syncing them piecewise.

    Args:
        == Source and destination parameters ==
        sfs_name: Source filesystem name.
        sfs_opts: Source filesystem options, as passed to fs_<name> class.
        dfs_name: Destination filesystem name.
        dfs_opts: Dest. filesystem options, as passed to fs_<name> class.
        sourcepath: File path on source system to be replicated.
        destpath: File path on destination system to store replication.

        == Source selection parameters ==
        checksumfile: File path on source with checksums for all files
            under <sourcepath>. Wildcards are allowed.
        checksumfileformat: Format of checksum file, either:
            text: one line per file, consisting of cheksum and filename
            json: structured file following the jsonschema defined.
        checksumfileschema: Only relevant for ckecksumfileformat json: file containing a jsonschema definition 
            the checksumfile will be validated against
        compareChecksums: Flag to check file checksums syncing.
        flagfile: Filename, when non-empty this ignores (sub)folders on 
            source filesystem that do not contain such a file.
        flag_age: minimum age (in seconds) of flag file before starting import
        minage: Filter on minimum age of subfolder.
        maxage: Filter on maximum age of subfolder.
        pattern: Regex filter on path of subfolder.
        exclude: List of regex patterns of files to exclude from sync
        timestamp_list: dict of attr: age pairs. The destination collection must exist. All
            the _attr_ attributes must be present. The value is assumed to be a timestamp. In order to be copied,
            it has to be at least _age_ old. Added especially for adding image data to a collection after 
            processing has finished

        == uniqueid parameters == 
        The script adds uniqueid metadata to the destination folder.
        This is used to determine if a data import has already been started
        or finished earlier, and the destination folder has moved in the meantime.

        == Metadata parameters ==
        metadata: List of AVUs to be set on the destination folder on
            succesful replication. 

        == Misc parameters ==
        skip_subdirs: Integer, descend this number of single-entry 
            directories before starting sync.
        copy_procs: Integer, number of processes to use for parallel file
            transfer.
    """

    # The (optional) schema definition is global, and should be read and parsed only once,
    # checksumfileschema is set by default.
    if checksumfileformat in [FILE_FORMAT_BASECLEAR, FILE_FORMAT_GENERIC_JSON]:
        with open(checksumfileschema) as f:
            JSON_SCHEMA = json.load(f)
    else:
        JSON_SCHEMA = None

    # Check all the remote data folders for first level subfolders
    logger.debug('Replicate %s to %s' % (sourcepath, destpath))

    fs_source = factory.createfs(sfs_name, **sfs_opts)
    fs_dest = factory.createfs(dfs_name, **dfs_opts)

    """
    Skip skip_subdirs subdirs, to handle minion data export
    """
    def folderlist(fs, current_dir, levels=0):
        if levels == 0:
            try:
                dirs = fs.lsdirs(current_dir, skip_inaccessible=True)
            except:
                # Skip inaccessible/special directories
                logger.error("ERROR reading directory %s" % current_dir)
                dirs = []
        else:
            dirs = []
            for subfolder in fs.lsdirs(current_dir, skip_inaccessible=True):
                dirs = dirs + folderlist(fs, subfolder.path, levels-1)
        return dirs

    for folder in folderlist(fs_source, sourcepath, levels=skip_subdirs):
        logger.debug("Process folder %s" % folder.path)
        if pattern:
            rePattern = re.compile(pattern + '$', re.IGNORECASE)
            if not rePattern.match(folder.shortname()):
                logger.debug('Folder %s does not match pattern %s' %
                             (folder.shortname(), pattern))
                continue
        if flagfile:
            # Note: flagfile may contain wildcards!
            # List files matching the 'flagfile' pattern,
            # determine the age of the newest one
            # and check against flag_age
            myflagfile = os.path.join(folder.path, flagfile)
            flagfile_list = fs_source.glob(myflagfile)
            min_age = 1E10
            if not flagfile_list:
                logger.debug('Cannot find flagfile %s' % myflagfile)
                continue
            for ff in flagfile_list:
                age = int(time.time() - fs_source.getfile(ff).utc_mtime())
                min_age = min(min_age, age)
            if min_age < flag_age:
                logger.info('Flag file(s) too new. Age is {}, min age is {}'.format(
                    min_age, flag_age))
                continue

        if minage > 0 or maxage < 999999999:
            logger.debug('Checking folder age {} - {}'.format(minage, maxage))
            folderage = int(time.time()) - fs_source.foldermtime(folder.path)
            logger.debug('Remote folder age is {} s'.format(folderage))
            if minage > 0:
                if folderage < minage:
                    logger.info('Recent changes - skipping folder')
                    continue
            if folderage > maxage:
                logger.info('Folder too old : %s - skipping' %
                            formattime(folderage))
                continue

        cslist = {}
        if checksumfile:
            cslist = parse_checksum_file(
                fs_source, folder, checksumfile, checksumfileformat, JSON_SCHEMA)
            if not cslist:
                continue

        destfolder = os.path.join(destpath, folder.shortname())
        if not fs_dest.folderexists(destfolder):
            fs_dest.mkdir(destfolder)

        objdfolder = fs_dest.getfolder(destfolder)
        sync_this_folder = True

        # Check timestamps on destination folder:
        if timestamp_list:
            logger.debug(
                'Verifying required timestamps on {}'.format(objdfolder.path))
            now = int(time.time())
            for t in timestamp_list:
                try:
                    v_str = objdfolder.getmeta(t)
                    v = float(v_str)
                except KeyError:
                    logger.error('timestamp attr {} missing on folder {}'.format(
                        t, objdfolder.path))
                    sync_this_folder = False
                    break
                except ValueError:
                    logger.error('timestamp attr {} on folder {} has invalid value {}'.format(
                        t, objdfolder.path, v_str))
                    sync_this_folder = False
                    break
                if now < (int(v) + int(timestamp_list[t])):
                    logger.info('timestamp attr {} on folder {} too new. Wait {} seconds'.format(
                        t, objdfolder.path, int(v)+int(timestamp_list[t])-now))
                    sync_this_folder = False
                    break

        # Params completion_avu determines copy selection
        # behaviour:
        # completion_avu == None: no check for completion attribute will be done
        #   when selecting folders to copy
        # completion_avu != None:
        #   only folders where completion_avu is not present will be copied

        if completion_avu and sync_this_folder:
            sync_this_folder = not test_completion_avu(
                objdfolder, completion_avu)
            if not sync_this_folder:
                logger.debug(
                    'Import for folder %s already marked complete' % folder.path)

        # Load the content of the filter files
        cs_filters = []
        if cs_filter_file:
            with open(cs_filter_file, 'r') as f:
                data = yaml.safe_load(f)
                cs_filters = data.get("filter", [])

        scan_filters = []
        if scan_filter_file:
            with open(scan_filter_file, 'r') as f:
                data = yaml.safe_load(f)
                scan_filters = data.get("filter", [])

        logger.info(f"Sync this folder: {folder.path}")

        if sync_this_folder:
            sync_folder(fs_source, folder, fs_dest, destfolder, objdfolder,
                        sfs_name, sfs_opts, dfs_name, dfs_opts,
                        compareChecksums,
                        cslist,
                        metadata,
                        copy_procs,
                        excludelist,
                        compare,
                        minimum_age,
                        cs_filters,
                        scan_filters,
                        scan)
            # SYNC might take a long time:
            # so refresh the connections
            fs_source.refresh()
            fs_dest.refresh()


def replicate_single_folder(sfs_name, sfs_opts, dfs_name, dfs_opts,
                            sourcepath, destpath,
                            compareChecksums=True, checksumfile='', checksumfileformat='', checksumfileschema='',
                            metadata={},
                            copy_procs=1,
                            excludelist=[],
                            completion_avu=None,
                            mark_complete=False,
                            compare=True,
                            timestamp_list=[],
                            minimum_age=None,
                            scan=False,
                            cs_filter_file='',
                            scan_filter_file=''):

    fs_source = factory.createfs(sfs_name, **sfs_opts)
    fs_dest = factory.createfs('irods', **dfs_opts)
    folder = fs_source.getfolder(sourcepath)
    destfolder = destpath
    objdfolder = None
    try:
        objdfolder = fs_dest.getfolder(destfolder)
    except ex.CollectionDoesNotExist:
        objdfolder = fs_dest.mkdir(destfolder)

    if completion_avu:
        if test_completion_avu(objdfolder, completion_avu):
            logger.debug(
                'Import for folder %s already marked complete' % sourcepath)
            return

    if checksumfileformat in [FILE_FORMAT_BASECLEAR, FILE_FORMAT_GENERIC_JSON]:
        with open(checksumfileschema) as f:
            JSON_SCHEMA = json.load(f)
    else:
        JSON_SCHEMA = None

    cslist = {}
    if checksumfile:
        cslist = parse_checksum_file(
            fs_source, folder, checksumfile, checksumfileformat, JSON_SCHEMA)
        if not cslist:
            logger.error('Given checksum file is empty')
            sys.exit(0)

    cs_filters = []
    if cs_filter_file:
        with open(cs_filter_file, 'r') as f:
            data = yaml.safe_load(f)
            cs_filters = data.get("filter", [])

    scan_filters = []
    if scan_filter_file:
        with open(scan_filter_file, 'r') as f:
            data = yaml.safe_load(f)
            scan_filters = data.get("filter", [])
    success = sync_folder(fs_source, folder, fs_dest, destfolder, objdfolder,
                          sfs_name, sfs_opts,
                          'irods', dfs_opts,
                          compareChecksums,
                          cslist,
                          metadata,
                          copy_procs,
                          excludelist,
                          compare,
                          minimum_age,
                          cs_filters,
                          scan_filters,
                          scan)
    if success:
        sys.exit(0)
    sys.exit(1)


def loglevel_convert(level):
    """Convert levels 1..5 to logging constants
    """
    LEVELS = dict(enumerate(
        [logging.CRITICAL, logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]))
    return LEVELS.get(min(max(1, level), len(LEVELS)) - 1, logging.WARNING)


def main():
    """
    Main function shows usage
    options used: -A -a -C -c -cf -cs -d -F -f -g -G -i -k -l -m -M -n -O -o -P -p -q -R -S -s -T -t -X -x -z
    """

    pre = argparse.ArgumentParser(add_help=False)
    pre.add_argument('-l', '--listfs', required=False,
                     action='store_true', default=False)
    args_pre = pre.parse_known_args()

    if args_pre[0].listfs:
        factory.listfs()
        exit(0)

    parser = argparse.ArgumentParser(description='Data synchronisation between iRODS and '
                                     'other file protocols.')

    parser.add_argument('source_path', help='Source path', type=str)
    parser.add_argument('coll', help='destination collection', type=str)

    # Source options
    parser.add_argument('-s', '--source_fs',
                        help='Source filesystem', default='local')
    parser.add_argument('-o', '--source_options', help='Source options')

    # Destination options
    parser.add_argument('-d', '--irods_options', help='iRods options')
    parser.add_argument('-R', '--resource',
                        help='Destination resource', default='')
    parser.add_argument('-S', '--ssl', help='Use SSL', action="store_true")

    # Source selection
    parser.add_argument('--search', help='Search for directories under the supplied path',
                        action="store_true")
    parser.add_argument('-n', '--skip_subdirs',
                        help='Skip this number of subdirectories while searching for source directories', type=int, default=0)
    parser.add_argument('-P', '--folder_pattern',
                        help='Pattern to match foldername', default='')
    parser.add_argument('-f', '--flag_filename',
                        help='Flag filename', default='')
    parser.add_argument('-g', '--flag_age',
                        help='Flag filename min age', type=int, default=0)
    parser.add_argument('-c', '--checksum_file',
                        help='Checksum file. Wildcards are allowed.', default='')
    parser.add_argument('-cf', '--checksum_file_format', help='Checksum file format ("%s"/"%s"/"%s")' %
                        (FILE_FORMAT_TEXT, FILE_FORMAT_BASECLEAR, FILE_FORMAT_GENERIC_JSON), default=FILE_FORMAT_TEXT)
    parser.add_argument('-cs', '--checksum_file_schema',
                        help='file containing JSONSCHEMA definition of checksum file format.', default=GENERIC_JSON_SCHEMA_FILE)

    # include the --filter_file alias for backward compatibility
    parser.add_argument('--checksum_filter_file', '--filter_file',
                        help='A yaml-file containing filter rules to include/exclude files/directories in the checksum_file', default='')
    parser.add_argument(
        '--scan', help='Scan (walk) source dir to find files to copy. This defaults to true if -c is not supplied', action='store_true')
    parser.add_argument(
        '--scan_filter_file', help='A yaml-file containing filter rules to include/exclude files/directories found by scanning the source', default='')
    parser.add_argument(
        '-a', '--minage', help='Minimum folder age in seconds', type=int, default=0)
    parser.add_argument(
        '-A', '--maxage', help='Maximum folder age in seconds', type=int, default=999999999)
    parser.add_argument('-X', '--exclude', help='Relative file/dir pattern to exclude',
                        type=str, action="append", default=[])
    parser.add_argument('-T', '--timestamp_age',
                        help='Supply attr=age. Checks for a timestamp attr on dest coll, that is min age old', default=[], action="append")
    parser.add_argument(
        '-w', '--last_write', help='Seconds since last writing access to file (in seconds).', type=int, default=0)

    # Metadata options
    parser.add_argument(
        '-m', '--metadata', help='Metadata to add to destination collection', default=[], action='append')
    parser.add_argument(
        '-M', '--metadata-file', help='JSON file with metadata to add to destination collection', default=None)
    parser.add_argument('-O', '--completion_avu',
                        help='Metadata attr=value to check for to skip folders', default=None)

    # Copy options
    parser.add_argument('-x', '--verify_checksums',
                        help='Verify checksums after copy', action="store_true")
    parser.add_argument(
        '-z', '--no_compare', help='Do not compare existing objects by size/date/checksum', action="store_true")
    parser.add_argument('-t', '--copy_procs',
                        help='Number of copy processes', type=int, default=1)

    # Logging options
    parser.add_argument('--data_source_name',
                        help='Name of data source', default='intorods')
    parser.add_argument('--debuglevel', help='Debuglevel (1..5)',
                        type=int, default=4)
    parser.add_argument('--syslog', help='Log to syslog', action='store_true')

    args = parser.parse_args()

    assert args.source_path, "Source path is required"
    assert args.coll, "Destination collection is required"

    loglevel = loglevel_convert(args.debuglevel)

    log_to_stdout(logger, args.data_source_name, loglevel)
    if args.syslog:
        log_to_syslog(logger, args.data_source_name, loglevel)

    logger.debug('{} startup for data source "{}".'.format(
        __name__, args.data_source_name))

    source_options = parse_extra_options(args.source_options)

    metadata = {} 
    if args.metadata_file:
        fs_source = factory.createfs(args.source_fs, **source_options)
        metadata_file = fs_source.getfile(args.metadata_file)
        if not metadata_file.accessible():
            logger.error('metadata-file %s is not accessible' % args.metadata_file)
            exit( 1 )
        with metadata_file.open('r') as fh:
            metadata_file_content = json.load(fh)
            metadata.update(metadata_file_content)
    metadata.update(dict(m.split('=') for m in args.metadata))

    dest_options = parse_extra_options(args.irods_options)
    dest_options.update({'resource': args.resource, 'use_ssl': args.ssl})

    timestamp_list = {opt.split('=')[0]: opt.split('=')[1]
                      for opt in args.timestamp_age}

    # Using scan defaults to True if no filelist is specified.
    # This ensures backward compatibility
    scan = True if not args.checksum_file else args.scan

    if args.search:
        replicate_data_folder(args.source_fs, source_options, 'irods', dest_options,
                              args.source_path, args.coll,
                              compareChecksums=args.verify_checksums,
                              minage=args.minage, maxage=args.maxage,
                              flagfile=args.flag_filename,
                              flag_age=args.flag_age,
                              checksumfile=args.checksum_file,
                              checksumfileformat=args.checksum_file_format,
                              checksumfileschema=args.checksum_file_schema,
                              pattern=args.folder_pattern,
                              metadata=metadata,
                              skip_subdirs=args.skip_subdirs,
                              copy_procs=args.copy_procs,
                              excludelist=args.exclude,
                              completion_avu=args.completion_avu,
                              compare=not args.no_compare,
                              timestamp_list=timestamp_list,
                              minimum_age=args.last_write,
                              cs_filter_file=args.checksum_filter_file,
                              scan=scan,
                              scan_filter_file=args.scan_filter_file
                              )
    else:
        if args.coll[-1] == '/':
            destination_coll = os.path.join(
                args.coll, os.path.basename(args.source_path))
        else:
            destination_coll = args.coll
        replicate_single_folder(args.source_fs, source_options, 'irods', dest_options,
                                args.source_path, destination_coll,
                                compareChecksums=args.verify_checksums,
                                checksumfile=args.checksum_file,
                                checksumfileformat=args.checksum_file_format,
                                checksumfileschema=args.checksum_file_schema,
                                metadata=metadata,
                                copy_procs=args.copy_procs,
                                excludelist=args.exclude,
                                completion_avu=args.completion_avu,
                                compare=not args.no_compare,
                                minimum_age=args.last_write,
                                scan_filter_file=args.scan_filter_file,
                                cs_filter_file=args.checksum_filter_file,
                                scan=scan
                                )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan  8 11:38:22 2020

@author: Erwin van Wieringen
"""

import argparse
import json
import hashlib
import logging
import logging.config
import os
import sys
import time
from datetime import datetime
from multiprocessing import Pool

BUF_SIZE = 1024 * 1024

log_format = 'hash: %(asctime)s|%(name)s %(levelname)s: %(message)s'

def checksum(fileobject):
    logging.debug(f'Hashing {fileobject[1]}')
    file = fileobject[1]
    sha256 = hashlib.sha256()
    with open(file, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha256.update(data)
    return (fileobject[0], sha256.hexdigest())

def checksum2(file):
    sha256 = hashlib.sha256()
    with open(file, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha256.update(data)
    return sha256.hexdigest()


def create_hash(folder, procs=1):
    logging.debug('Generate hash jobs')
    hashjobs = []

    basedir = os.path.normpath(folder)
    for root, dirs, files in os.walk(basedir):
        reldir = root[len(basedir)+1:]
        for file in files:
            hashjobs.append((os.path.join(reldir, file), os.path.join(root, file)))
    logging.info(f'{len(hashjobs)} files to hash')
    logging.info(f'Start {procs} hash workers')
    pool = Pool(processes=procs)
    hashlist = pool.map(checksum, hashjobs)    
    hashes = { item[0]: item[1] for item in hashlist}
    logging.info('Hashing finished')        
    return hashes

def loglevel_convert(level):
    """Convert levels 1..5 to logging constants
    """
    LEVELS = dict(enumerate([logging.CRITICAL, logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]))
    return LEVELS.get(level-1, logging.WARNING)

def hash_dict_to_json(hashdict, collection=""):
    objects = []
    for path, hash in hashdict.items():
        file_json = {
            "path": path,
            "checksum": hash,
            "type": "dataobject"
        }
        objects.append(file_json)

    result_json = {
        "collection": collection,
        "checksum_format": "sha256",
        "version": 1,
        "checksum_encoding": "hex",
        "objects": objects
    }
    return json.dumps(result_json, indent=3)

def hash_dict_to_txt(hashdict):
    output = ""
    for path, hash in hashdict.items():
        output += "{}  {}\n".format(hash, path)
    return output

def cmd_parser():
    parser = argparse.ArgumentParser(description='Hash file generator')
    parser.add_argument('sourcedir', help='Source directory')
    parser.add_argument('-o', '--output', help='Output file')
    parser.add_argument('-j', '--json', help='Output in json format', action='store_true')
    parser.add_argument('-p', '--procs', help='Hash processes', type=int, default=1)
    parser.add_argument('-d', '--debuglevel', help='Debug level (1-5)', type=int, default=logging.CRITICAL)
    parser.add_argument('-C', '--coll', help='destination collection for use in json output', default='')

    return parser.parse_args()

def main():
    """
    Main function shows usage
    """

    args = cmd_parser()

    logging.basicConfig(format=log_format, level=loglevel_convert(args.debuglevel))

    t0 = time.time()

    hashes = create_hash(args.sourcedir, procs=args.procs)

    if args.output is not None:
        hashes.pop(args.output, None)
    
    hashtime = time.time() - t0

    
    output = ""
    if args.json:
        output = hash_dict_to_json(hashes, collection=args.coll)
    else:
        output = hash_dict_to_txt(hashes)

    if args.output:
        hashfile = open(args.output, 'w')
    else:
        hashfile = sys.stdout
    hashfile.write(output)
    hashfile.close

    logging.info(f'Hashed {len(hashes)} files in {hashtime:.2f} seconds')

if __name__ == "__main__":
    main()
    


import json
import os
import re


# returns number of non-empty path components
def path_length( path ):
    # '/' = 0
    # '/A/' = 1
    parts = path.split(os.sep)
    l=0
    for p in parts:
        if p:
            l += 1
    return l

class PathFilter:

    def __init__(self, filter):
        def addFilterDef( filter ):
            parts = filter.split(' ')
            if len( parts ) not in [2,3]:
                raise f"ERROR: invalid filter: '{filter}'"
            include = parts[0] == '+'
            if os.path.basename( parts[1] ):
                raise f"ERROR: invalid filter: '{parts}'. All filter must end in '/'"
            fpath = os.path.normpath( os.path.dirname( parts[1] ))
            #its a dir filter
            if len( parts ) == 2:
                entry = (include, fpath)
                self._dir_filter.append( entry )
                return
            #its a file filter
            regex = re.compile(parts[2], re.IGNORECASE) 
            entry = (include, fpath, regex)
            if include:
                self._file_filter_include.setdefault( fpath, [] ).append( entry )
            else:
                self._file_filter_exclude.setdefault( fpath, [] ).append( entry )

        #filter rules from yml-file...
        #...regarding directories
        self._dir_filter = []
        #...regarding files
        self._file_filter_include = {}
        self._file_filter_exclude = {}

        for f in filter:
            addFilterDef(f)

    def show( self ):
        print("Filter:")
        for f in self._dir_filter:
            print(f)
        for k in self._file_filter_include.items():
            print(k)
            print(f)
        for k,f in self._file_filter_exclude.items():
            print(k)
            print(f)
        print("------")

    def is_a_subpath( self, cand, path):
        score = None    
        ncand = os.path.normpath( cand )
        if ncand == path:
            #its perfect match
            score = 0
        elif ncand.startswith( '{}/'.format(path) ):
            #its a subpath!
            score = path_length( ncand ) - path_length(path)
        elif path == "/":
            #by default everything is decended of /
            score = path_length( ncand )
        return score


    def get_nearest_directory_filter( self, path ):
        scores = []
        for entry in self._dir_filter:
            include = entry[0]
            score = self.is_a_subpath( path, entry[1])
            if score != None:
                scores.append( (score, entry) )
        #find the rule with highest score:
        if not scores:
            return None
        min_score = min( scores, key=lambda score: score[0])
        return min_score[1]


    # /path/to/directory
    def isDirIncluded( self, path ):
        entry = self.get_nearest_directory_filter( path )
        include=True
        if entry:
            include = entry[0]
        return include

    # /path/to/filename.txt
    def isFileIncluded( self, path_and_file ):
        ncand = os.path.normpath( path_and_file )
        path, filename = os.path.split( path_and_file )
        if not path:
            path = os.sep
        if not filename:
            print( "WARN: path '{path_and_file}' doesn't contain a filename" )
            return False
        
        inFilter = self._file_filter_include.get( path, [] )
        exFilter = self._file_filter_exclude.get( path, [] )

        #if there is any inclusion filter, then by default all files are excluded...
        includeFile = self.isDirIncluded( path )
        matches = 0
        for entry in inFilter:
            if entry[2].match(filename):
                matches += 1
                #print(entry)
                includeFile = True
        for entry in exFilter:
            if entry[2].match(filename):
                matches += 1
                #print(entry)
                includeFile = False
        if matches > 1:
            print( f"WARN: multiple matches for path: '{path_and_file}'")
        return includeFile




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May  7 12:14:23 2019

@author: Erwin van Wieringen
"""

import base64
import calendar as cal
import logging
import math
import os
import re
import ssl
import time

import irods.keywords as kw
from irods.column import Criterion
from irods.models import Collection, CollectionMeta, DataObject
from irods.session import iRODSSession

from intorods.filesys.fs_base import factory, fs_base, fsobject_base

logger = logging.getLogger(__name__)


def utc2local(utc):
    return time.mktime(time.localtime(utc))


class file_irods(fsobject_base):
    def __init__(self, fso, path):
        super().__init__(fso, path)
        self.refresh()

    def refresh(self):
        self.irods_object = self.fso.irods_session.data_objects.get(self.path)

    def isdir(self):
        return False

    def calculate_checksum(self):
        cs = self.irods_object.checksum
        if cs is None:
            return super().calculate_checksum()
        else:
            sha2 = cs.split(':')[1]
            return base64.b64decode(sha2).hex()

    def filesize(self):
        return self.irods_object.size

    def isfile(self):
        return True

    def open(self, mode):
        return self.irods_object.open(mode[:1])

    def set_mtime(self, mtime):
        new_time = utc2local(mtime)
        self.irods_object.manager.modDataObjMeta(
            {"objPath": self.path}, {"dataModify": round(new_time)})
        self.refresh()

    def utc_mtime(self):
        mtime = cal.timegm(self.irods_object.modify_time.timetuple())
        return math.floor(mtime)

# TODO
    def local_mtime(self):
        return 0


class folder_irods(fsobject_base):
    def __init__(self, fso, path):
        super().__init__(fso, path)
        self.irods_object = fso.irods_session.collections.get(path)

    def filesize(self):
        return 0

    def isdir(self):
        return True

    def isfile(self):
        return False

    def accessible(self):
        raise NotImplementedError

    def removemeta(self, name):
        for m in self.irods_object.metadata.get_all(name):
            self.irods_object.metadata.remove(m)

    def setmeta(self, name, value, units=''):
        try:
            self.irods_object.metadata.add(name, value, units)
        except:
            pass

    def replaceorsetmeta(self, name, value):
        self.removemeta(name)
        self.setmeta(name, value)

    def getmeta(self, name):
        return self.irods_object.metadata.get_one(name).value

    def getorsetmeta(self, name, default):
        if not self.irods_object.metadata.get_all(name):
            self.setmeta(name, default, '')
        return self.getmeta(name)

    def utc_mtime(self):
        return 0


class fs_irods(fs_base):
    def __init__(self, resource='', use_ssl=False, host=None, user=None,
                 zone=None, password=None, port=1247, timeout=None):
        super().__init__(supportsopen=True)

        if host:
            session_params = {'host': host, 'port': port, 'zone': zone,
                              'user': user, 'password': password}
        else:
            try:
                env_file = os.environ['IRODS_ENVIRONMENT_FILE']
            except KeyError:
                env_file = os.path.expanduser(
                    '~/.irods/irods_environment.json')
            session_params = {'irods_env_file': env_file}

        if use_ssl:
            context = ssl._create_unverified_context(purpose=ssl.Purpose.SERVER_AUTH,
                                                     cafile=None, capath=None, cadata=None)
            ssl_settings = {'irods_ssl_ca_certificate_file': '/etc/irods/ssl/irods.crt',
                            'ssl_context': context}
            self.irods_session = iRODSSession(**session_params, **ssl_settings)
        else:
            self.irods_session = iRODSSession(**session_params)
        if timeout:
            self.irods_session.connection_timeout = int(timeout)
        self.resource = resource

    def cleanup(self):
        if self.irods_session:
            self.irods_session.cleanup()
            self.irods_session = None

    def __del__(self):
        self.cleanup()

    def ls(self, path):
        result = []
        coll = self.irods_session.collections.get(path)
        for subcoll in coll.subcollections:
            result.append(folder_irods(self, subcoll.path))
        for entry in coll.data_objects:
            result.append(self.getfile(entry.path))
        return result

    def lsdirs(self, path, skip_inaccessible=False):
        logger.error("lsdirs not implemented")
        exit(2)

    @staticmethod
    def factory(**kwargs):
        return fs_irods(**kwargs)

    def fileexists(self, path):
        base, file = self._pathsplit(path)
        query = self.irods_session.query(
            Collection.name, DataObject.name).filter(
                Criterion('=', Collection.name, base)).filter(
                    Criterion('=', DataObject.name, file))
        return bool(list(query.get_results()))

    def findfolder(self, name, meta=[]):
        result = []
        query = self.irods_session.query(Collection.name).filter(
            Criterion('like', Collection.name, '%/' + name))
        for m in meta:
            query = query.filter(
                Criterion('=', CollectionMeta.name, m['field'])).filter(
                    Criterion(m['op'], CollectionMeta.value, m['value']))
        for coll in query.get_results():
            collname = coll[Collection.name]
            if not re.match('/[^/]*/trash/.*', collname):
                result.append(folder_irods(self, collname))
        return result

    def folderexists(self, path):
        query = self.irods_session.query(Collection).filter(
            Criterion('=', Collection.name, path))
        return bool(list(query.get_results()))

    def _getfile(self, path):
        return file_irods(self, path)

    def getfolder(self, path):
        return folder_irods(self, path)

    def mkdir(self, path, parents=False):
        if parents:
            parentdir, dir = os.path.split(path)
            if not self.folderexists(parentdir):
                self.mkdir(parentdir, parents=True)
        self.irods_session.collections.create(path)
        return folder_irods(self, path)

# Opening an object with create=True should be used for irods4.3.0
# instead of create and open in two statements.
# Check: https://github.com/irods/irods/issues/6808
    def createfile(self, path):
        options = {kw.DEST_RESC_NAME_KW: self.resource}
        return self.irods_session.data_objects.open(path, 'w', create=True, **options)

    def deletefile(self, path):
        self.irods_session.data_objects.unlink(path, True)
        self.invalidate_cache_entry(path)

    def open(self, path, mode):
        options = {kw.FORCE_FLAG_KW: ''}
        obj = self.irods_session.data_objects.create(path, **options,
                                                     resource=self.resource)
        return obj.open(mode[:1])


factory.register('irods', fs_irods.factory,
                 'resource=<dest resource>,use_ssl=<true|false>,host=<host>,user=<user>,password=<passwd>,zone=<zone>,timeout=<timeout>')



import os
import sys

import pytest

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from intorods.intorods import formattime, parse_extra_options


@pytest.mark.parametrize("ts, output",[(1,' 0 days  0: 0: 1 '),(1000,' 0 days  0:16:40 '),(100000,' 1 days  3:46:40 '),(10000000,'115 days 17:46:40 ')])
def test_formattime(ts, output):
    result = formattime(ts)
    assert result == output

def test_parse_extra_options():
    result = parse_extra_options('option1=value1,option2=value2')
    assert result == { 'option1': 'value1', 'option2': 'value2'}




import logging
import logging.handlers
import sys
from logging.handlers import SysLogHandler

LOG_FORMAT_FILE = "%(asctime)s|%(name)s %(levelname)s: %(message)s"
LOG_PARAMS = {
    "when": "D",
    "interval": 5,
    "backupCount": 20
}


def log_to_file(logger, filename, log_level):
    logger.setLevel(log_level)
    formatter = logging.Formatter(LOG_FORMAT_FILE)
    fh = logging.handlers.TimedRotatingFileHandler(filename, **LOG_PARAMS)
    fh.setFormatter(formatter)
    fh.setLevel(log_level)
    logger.addHandler(fh)


def log_to_syslog(logger, name, log_level):
    logger.setLevel(log_level)
    log_format = f'{name}: %(asctime)s|%(name)s %(levelname)s: %(message)s'
    formatter = logging.Formatter(log_format)
    lg = logging.handlers.SysLogHandler(
        facility=SysLogHandler.LOG_LOCAL6, address='/dev/log')
    lg.setFormatter(formatter)
    lg.setLevel(log_level)
    logger.addHandler(lg)


def log_to_stdout(logger, name, log_level):
    logger.setLevel(log_level)
    log_format = f'{name}: %(asctime)s|%(name)s %(levelname)s: %(message)s'
    formatter = logging.Formatter(log_format)
    lg = logging.StreamHandler(stream=sys.stdout)
    lg.setFormatter(formatter)
    lg.setLevel(log_level)
    logger.addHandler(lg)



# -*- coding: utf-8 -*-
"""
Created on Fri Jan 17 09:12:30 2020
SMB class for the filesys library
@author: Erwin van Wieringen
"""
import logging
import socket
import tempfile
import os

from smb.SMBConnection import SMBConnection
from smb import smb_constants

from intorods.filesys.fs_base import factory, fs_base, fsobject_base

logger = logging.getLogger(__name__)


def myshortname():
    return socket.gethostname().split('.')[0]


def uxtowin(path):
    return '\\'.join(path.split('/'))


def wintoux(path):
    return '/'.join(path.split('\\'))


class file_smb(fsobject_base):
    def __init__(self, fso, path):
        super().__init__(fso, path)
        self.attr = fso.smb_conn.getAttributes(
            self.fso.share, uxtowin(self.path))

    def copyto(self, dest_fs, dest_path):
        with dest_fs.createfile(dest_path) as destfile:
            self.fso.smb_conn.retrieveFile(
                self.fso.share, uxtowin(self.path), destfile)

    def close(self):
        self.file_obj.close()

    def open(self, mode):
        self.file_obj = tempfile.NamedTemporaryFile()
        self.fso.smb_conn.retrieveFile(
            self.fso.share, uxtowin(self.path), self.file_obj)
        self.file_obj.seek(0)
        return self.file_obj

    def filesize(self):
        return self.attr.file_size

    def isdir(self):
        return self.attr.file_attributes & smb_constants.ATTR_DIRECTORY > 0

    def isfile(self):
        return not self.isdir()

    def utc_mtime(self):
        return int(self.attr.last_write_time//1)

# TODO
    def local_mtime(self):
        return 0

# TODO
    def set_mtime(self, mtime):
        pass


class fs_smb(fs_base):
    def __init__(self, server='', share='', nbserver='', username='',
                 domain='', password=''):
        super().__init__(supportsopen=False)
        self.server = server
        if not nbserver:
            nbserver = server.split('.')[0].upper()
        self.nbserver = nbserver
        self.username = username
        self.password = password
        self.domain = domain
        self.smb_conn = None
        self.connect()
        self.share = share

    def connect(self):
        client_machine_name = myshortname()
        if self.smb_conn:
            self.smb_conn.close()
        self.smb_conn = SMBConnection(self.username, self.password, client_machine_name,
                                      self.nbserver, domain=self.domain, is_direct_tcp=True)
        if not self.smb_conn.connect(self.server, 445):
            logger.error('Connection to %s failed' % self.server)
            exit(2)

    def refresh(self):
        self.connect()

    def copyfrom(self, source, path):
        self.smb_conn.storeFile(self.share, uxtowin(path), source)

    @staticmethod
    def factory(**kwargs):
        return fs_smb(**kwargs)

    def fileexists(self, path):
        try:
            _ = self.smb_conn.getAttributes(self.share, uxtowin(path))
            return True
        except:
            return False

    def folderexists(self, path):
        path_exists = self.exists(path)
        if path_exists:
            dirobject = self.getfile(path)
            return dirobject.isdir()
        return False

    def _getfile(self, path):
        return file_smb(self, path)

    def lsdirs(self, path, skip_inaccessible=False):
        result = []
        for entry in self.smb_conn.listPath(self.share, uxtowin(path),
                                            search=smb_constants.SMB_FILE_ATTRIBUTE_DIRECTORY):
            if not entry.filename in ['.', '..']:
                fullpath = path + '/' + entry.filename
                try:
                    fileobject = self.getfile(fullpath)
                    result.append(fileobject)
                except:
                    pass
        return result

    def lsdirnames(self, path):
        result = []
        for entry in self.smb_conn.listPath(self.share, uxtowin(path),
                                            search=smb_constants.SMB_FILE_ATTRIBUTE_DIRECTORY):
            if not entry.filename in ['.', '..']:
                result.append(entry.filename)
        return result

    def lsfilenames(self, path):
        result = []
        SEARCH_MASK = smb_constants.SMB_FILE_ATTRIBUTE_HIDDEN | \
            smb_constants.SMB_FILE_ATTRIBUTE_SYSTEM | \
            smb_constants.SMB_FILE_ATTRIBUTE_ARCHIVE | \
            smb_constants.SMB_FILE_ATTRIBUTE_INCL_NORMAL
        for entry in self.smb_conn.listPath(self.share, uxtowin(path),
                                            search=SEARCH_MASK):
            result.append(entry.filename)
        return result

    def mkdir(self, path, parents=False):
        if parents:
            parentdir, dir = os.path.split(path)
            if not self.folderexists(parentdir):
                self.mkdir(parentdir, parents=True)
        self.smb_conn.createDirectory(self.share, uxtowin(path))

# TODO
    def open(self, path, mode):
        return None

    def ls(self, path):
        result = []
        for entry in self.smb_conn.listPath(self.share, uxtowin(path)):
            fullpath = path + '/' + entry.filename
            fileobject = self.getfile(fullpath)
            result.append(fileobject)
        return result


factory.register('smb', fs_smb.factory,
                 'server=<server>, share=<sharename>, \
                 nbserver=<netbios servername>, username=<user>, \
                 domain=<domain>, password=<password>')
